/*
 *  Copyright 2020-2024 Yuntu Microelectronics co.,ltd
 *  All rights reserved.
 * 
 *  YUNTU Confidential. This software is owned or controlled by YUNTU and may only be
 *  used strictly in accordance with the applicable license terms. By expressly
 *  accepting such terms or by downloading, installing, activating and/or otherwise
 *  using the software, you are agreeing that you have read, and that you agree to
 *  comply with and are bound by, such license terms. If you do not agree to be
 *  bound by the applicable license terms, then you may not retain, install,
 *  activate or otherwise use the software. The production use license in
 *  Section 2.3 is expressly granted for this software.
 * 
 * @file fee_config.c
 * @brief 
 * 
 */


#include <stddef.h>
#include "fee_config.h"


/* general */
static const Fls_SectorType sectorList[8]={
    {
        .sectorId=0,
        .sectorStartAddress=0x0,
        .sectorSize=0x200,
        .pageSize=4,
        .sectorHwStartAddress=0x10000000,
        .asyncAccess=0,
    },
    {
        .sectorId=1,
        .sectorStartAddress=0x200,
        .sectorSize=0x200,
        .pageSize=4,
        .sectorHwStartAddress=0x10000200,
        .asyncAccess=0,
    },
    {
        .sectorId=2,
        .sectorStartAddress=0x400,
        .sectorSize=0x200,
        .pageSize=4,
        .sectorHwStartAddress=0x10000400,
        .asyncAccess=0,
    },
    {
        .sectorId=3,
        .sectorStartAddress=0x600,
        .sectorSize=0x200,
        .pageSize=4,
        .sectorHwStartAddress=0x1f600,
        .asyncAccess=0,
    },
    {
        .sectorId=4,
        .sectorStartAddress=0x800,
        .sectorSize=0x200,
        .pageSize=4,
        .sectorHwStartAddress=0x1f800,
        .asyncAccess=0,
    },
    {
        .sectorId=5,
        .sectorStartAddress=0xa00,
        .sectorSize=0x200,
        .pageSize=4,
        .sectorHwStartAddress=0x1fa00,
        .asyncAccess=0,
    },
    {
        .sectorId=6,
        .sectorStartAddress=0xc00,
        .sectorSize=0x200,
        .pageSize=4,
        .sectorHwStartAddress=0x1fc00,
        .asyncAccess=0,
    },
    {
        .sectorId=7,
        .sectorStartAddress=0xe00,
        .sectorSize=0x200,
        .pageSize=4,
        .sectorHwStartAddress=0x1fe00,
        .asyncAccess=0,
    },
};

extern void Fee_JobEndNotification(void);
extern void Fee_JobErrorNotification(void);




static const Fls_ConfigType flashConfig={
    .acEraseFunPtr=NULL,
    .acWriteFunPtr=NULL,
    .jobEndNotificationFunPtr=Fee_JobEndNotification,
    .jobErrorNotificationFunPtr=Fee_JobErrorNotification,
    .eDefaultMode=MEMIF_MODE_SLOW,
    .maxReadFastMode=512,
    .maxReadNormalMode=256,
    .maxWriteFastMode=128,
    .maxWriteNormalMode=8,
    .ConfigSectorNum=8,
    .sectorList=sectorList,
};


/* cluster */
static Fee_ClusterType clrType0[4]={
    {
        .StartAddr=0x0,
        .Length=0x200,
    },
    {
        .StartAddr=0x200,
        .Length=0x200,
    },
    {
        .StartAddr=0x400,
        .Length=0x200,
    },
    {
        .StartAddr=0x600,
        .Length=0x200,
    },
};
static Fee_ClusterType clrType1[4]={
    {
        .StartAddr=0x800,
        .Length=0x200,
    },
    {
        .StartAddr=0xa00,
        .Length=0x200,
    },
    {
        .StartAddr=0xc00,
        .Length=0x200,
    },
    {
        .StartAddr=0xe00,
        .Length=0x200,
    },
};

static const Fee_ClusterGroupType clusterConfig[2]={
    {
        .ReservedSize=0,
        .ClrCount=4,
        .ClrPtr=clrType0,
    },
    {
        .ReservedSize=0,
        .ClrCount=4,
        .ClrPtr=clrType1,
    },
};

/* block */
static const Fee_BlockConfigType blockConfig[3]={
    {
        .BlockNumber=1,
        .BlockSize=4,
        .ClrGrp=0,
        .ImmediateData=0,
        .BlockAssignment=FEE_PROJECT_RESERVED,
    },
    {
        .BlockNumber=2,
        .BlockSize=50,
        .ClrGrp=0,
        .ImmediateData=0,
        .BlockAssignment=FEE_PROJECT_RESERVED,
    },
    {
        .BlockNumber=3,
        .BlockSize=16,
        .ClrGrp=1,
        .ImmediateData=0,
        .BlockAssignment=FEE_PROJECT_RESERVED,
    },
};



/* top */

const Fee_ModuleUserConfig_t FEEGenConfig={
    .blockCnt=3,
    .clusterCnt=2,
    .clusterConfigPtr=clusterConfig,
    .blockConfigPtr=blockConfig,
    .flashConfigPtr=&flashConfig,
};





